/*
 * ADBMS_Abstraction.c
 *
 *  Created on: 14.07.2022
 *      Author: max
 */

#include "ADBMS_Abstraction.h"
#include "ADBMS_CMD_MAKROS.h"
#include "ADBMS_LL_Driver.h"
#include "swo_log.h"
#include <stddef.h>

extern uint8_t numberofCells;

#define CHECK_RETURN(x)                                                        \
  {                                                                            \
    uint8_t status = x;                                                          \
    if (status != 0)                                                           \
      return status;                                                           \
  }

uint8_t amsReset() {
  amsWakeUp();
  readCMD(SRST, CMD_EMPTY_BUFFER, CMD_EMPTY_BUFFER_SIZE);

  uint8_t buffer[CMD_BUFFER_SIZE(SID_GROUP_SIZE)] = {0};
  CHECK_RETURN(readCMD(RDSID, buffer, CMD_BUFFER_SIZE(SID_GROUP_SIZE)));

  debug_log(LOG_LEVEL_INFO, "BMS reset complete\n");

  if (DEBUG_CHANNEL_ENABLED(LOG_LEVEL_INFO)) {
    debug_log(LOG_LEVEL_INFO, "Found IDs: ");
    for (size_t i = 0; i < N_BMS; i++) {
      uint64_t id = 0;
      for (size_t j = 0; j < SID_GROUP_SIZE; j++) {
        id |= buffer[BUFFER_BMS_OFFSET(i, SID_GROUP_SIZE) + j] << (j * 8);
      }
      debug_log_cont(LOG_LEVEL_INFO, "0x%llx ", id);
    }
    debug_log_cont(LOG_LEVEL_INFO, "\n");
  }

  mcuDelay(10);
  amsWakeUp();
  amsStopBalancing();
  amsConfigOverUnderVoltage(DEFAULT_OV, DEFAULT_UV);

  CHECK_RETURN(writeCMD(CLRFLAG, buffer, 6)); //clear flags,
  CHECK_RETURN(writeCMD(CLOVUV, buffer, 6));  //OVUV flags
  CHECK_RETURN(writeCMD(ADCV | ADCV_CONT | ADCV_RD, NULL, 0)); //start continuous cell voltage measurement with redundancy
  CHECK_RETURN(writeCMD(ADAX | ADAX_CONV_ALL, NULL, 0)); //start aux measurement

  return 0;
}

uint8_t initAMS(SPI_HandleTypeDef* hspi) {
  adbmsDriverInit(hspi);
  return amsReset();
}

uint8_t amsWakeUp() {
  uint8_t buffer[CMD_BUFFER_SIZE(CFG_GROUP_A_SIZE)] = {0};
  return readCMD(RDCFGA, buffer, CMD_BUFFER_SIZE(CFG_GROUP_A_SIZE));
}

uint8_t amsCellMeasurement(Cell_Module* module) {
  #warning check conversion counter to ensure that continous conversion has not been stopped
  #warning check for OW conditions: ADSV | ADSV_OW_0 / ADSV_OW_1
  return amsReadCellVoltages(module);
}

uint8_t amsAuxAndStatusMeasurement(Cell_Module* module) {
  uint8_t rxbuf[AUX_GROUP_A_SIZE] = {};

  CHECK_RETURN(readCMD(RDSTATC, rxbuf, STATUS_GROUP_C_SIZE));

  module->status.CS_FLT = rxbuf[0] | (rxbuf[1] << 8);
  module->status.CCTS   = rxbuf[2] | (rxbuf[3] << 8);
  module->status.VA_OV  = (rxbuf[4] >> 7) & 0x01;
  module->status.VA_UV  = (rxbuf[4] >> 6) & 0x01;
  module->status.VD_OV  = (rxbuf[4] >> 5) & 0x01;
  module->status.VD_UV  = (rxbuf[4] >> 4) & 0x01;
  module->status.CED    = (rxbuf[4] >> 3) & 0x01;
  module->status.CMED   = (rxbuf[4] >> 2) & 0x01;
  module->status.SED    = (rxbuf[4] >> 1) & 0x01;
  module->status.SMED   = (rxbuf[4] >> 0) & 0x01;
  module->status.VDEL   = (rxbuf[5] >> 7) & 0x01;
  module->status.VDE    = (rxbuf[5] >> 6) & 0x01;
  module->status.COMPARE= (rxbuf[5] >> 5) & 0x01;
  module->status.SPIFLT = (rxbuf[5] >> 4) & 0x01;
  module->status.SLEEP  = (rxbuf[5] >> 3) & 0x01;
  module->status.THSD   = (rxbuf[5] >> 2) & 0x01;
  module->status.TMODCHK= (rxbuf[5] >> 1) & 0x01;
  module->status.OSCCHK = (rxbuf[5] >> 0) & 0x01;

  if (pollCMD(PLAUX) == 0x0) { //TODO: check for SPI fault
    return 0; // aux ADC data not ready
  }

  CHECK_RETURN(readCMD(RDAUXA, rxbuf, AUX_GROUP_A_SIZE));

  module->auxVoltages[0] = mV_from_ADBMS6830(rxbuf[0] | (rxbuf[1] << 8));
  module->auxVoltages[1] = mV_from_ADBMS6830(rxbuf[2] | (rxbuf[3] << 8));
  module->auxVoltages[2] = mV_from_ADBMS6830(rxbuf[4] | (rxbuf[5] << 8));

  CHECK_RETURN(readCMD(RDAUXB, rxbuf, AUX_GROUP_A_SIZE));

  module->auxVoltages[3] = mV_from_ADBMS6830(rxbuf[0] | (rxbuf[1] << 8));
  module->auxVoltages[4] = mV_from_ADBMS6830(rxbuf[2] | (rxbuf[3] << 8));
  module->auxVoltages[5] = mV_from_ADBMS6830(rxbuf[4] | (rxbuf[5] << 8));

  CHECK_RETURN(readCMD(RDAUXC, rxbuf, AUX_GROUP_A_SIZE));

  module->auxVoltages[6] = mV_from_ADBMS6830(rxbuf[0] | (rxbuf[1] << 8));
  module->auxVoltages[7] = mV_from_ADBMS6830(rxbuf[2] | (rxbuf[3] << 8));
  module->auxVoltages[8] = mV_from_ADBMS6830(rxbuf[4] | (rxbuf[5] << 8));

  CHECK_RETURN(readCMD(RDAUXD, rxbuf, AUX_GROUP_A_SIZE));

  module->auxVoltages[9] = mV_from_ADBMS6830(rxbuf[0] | (rxbuf[1] << 8));

  uint8_t rxbuffer[STATUS_GROUP_A_SIZE];

  CHECK_RETURN(readCMD(RDSTATA, rxbuffer, STATUS_GROUP_A_SIZE));

  module->internalDieTemp = rxbuffer[2] | (rxbuffer[3] << 8);

  CHECK_RETURN(readCMD(RDSTATB, rxbuffer, STATUS_GROUP_B_SIZE));
  module->digitalSupplyVoltage = mV_from_ADBMS6830(rxbuffer[0] | (rxbuffer[1] << 8));
  module->analogSupplyVoltage  = mV_from_ADBMS6830(rxbuffer[2] | (rxbuffer[3] << 8));
  module->refVoltage           = mV_from_ADBMS6830(rxbuffer[4] | (rxbuffer[5] << 8));

  CHECK_RETURN(writeCMD(ADAX | ADAX_CONV_ALL, NULL, 0)); //start aux measurement for next cycle

  return 0;
}

uint8_t amsConfigBalancing(uint32_t channels, uint8_t dutyCycle) {
  uint8_t buffer_a[PWM_GROUP_A_SIZE] = {};
  uint8_t buffer_b[PWM_GROUP_B_SIZE] = {};
  CHECK_RETURN(readCMD(RDPWMA, buffer_a, CFG_GROUP_A_SIZE));
  CHECK_RETURN(readCMD(RDPWMB, buffer_b, CFG_GROUP_B_SIZE));

  if (dutyCycle > 0x0F) { // there are only 4 bits for duty cycle
    return 1;
  }

  #warning fixme

  for (size_t i = 0; i < 16; i += 2) {
    if (i < 12) { // cells 0, 1 are in regbuffer[0], cells 2, 3 in regbuffer[1], ...
      buffer_a[i / 2] = ((channels & (1 << (i + 1))) ? (dutyCycle << 4) : 0) |
                         ((channels & (1 << i)) ? dutyCycle : 0);
    } else {
      buffer_b[(i - 12) / 2] = ((channels & (1 << (i + 1))) ? (dutyCycle << 4) : 0) |
                                ((channels & (1 << i)) ? dutyCycle : 0);
    }
  }

  CHECK_RETURN(writeCMD(WRPWMA, buffer_a, CFG_GROUP_A_SIZE));
  CHECK_RETURN(writeCMD(WRPWMB, buffer_b, CFG_GROUP_B_SIZE));

  return 0;
}

uint8_t amsStartBalancing(uint8_t dutyCycle) { return writeCMD(UNMUTE, NULL, 0); }

uint8_t amsStopBalancing() { return writeCMD(MUTE, NULL, 0); }

uint8_t amsSelfTest() { return 0; }

uint8_t amsConfigOverUnderVoltage(uint16_t overVoltage, uint16_t underVoltage) {
  uint8_t buffer[CFG_GROUP_A_SIZE];

  if (underVoltage & 0xF000 || overVoltage & 0xF000) { // only 12 bits allowed
    return 1;
  }

  CHECK_RETURN(readCMD(RDCFGB, buffer, CFG_GROUP_A_SIZE));

  //UV
  buffer[0] = (uint8_t) (underVoltage & 0xFF);
  buffer[1] &= 0xF0;
  buffer[1] |= (uint8_t) ((underVoltage >> 8) & 0x0F);

  //OV
  buffer[1] &= 0x0F;
  buffer[1] |= (uint8_t) (overVoltage << 4);
  buffer[2] = (uint8_t) (overVoltage >> 4);

  return writeCMD(WRCFGB, buffer, CFG_GROUP_A_SIZE);
}

uint8_t amsCheckUnderOverVoltage(Cell_Module* module) {
  uint8_t regbuffer[STATUS_GROUP_D_SIZE];
  uint32_t ov_uv_data = 0;
  CHECK_RETURN(readCMD(RDSTATD, regbuffer, STATUS_GROUP_D_SIZE));
  ov_uv_data = (regbuffer[0] <<  0) | (regbuffer[1] <<  8) | 
               (regbuffer[2] << 16) | (regbuffer[3] << 24);

  module->overVoltage = 0;
  module->underVoltage = 0;
  
  for (size_t i = 0; i < numberofCells; i++) { // ov/uv flags are 1-bit flags for each cell C0UV, C0OV, C1UV, C1OV, ...
    module->underVoltage |= (ov_uv_data >> (i * 2)) & 0x01;
    module->overVoltage  |= (ov_uv_data >> (i * 2 + 1)) & 0x01;
  }

  return 0;
}

uint8_t amsClearFlag() {
  uint8_t buffer[6] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
  return writeCMD(CLRFLAG, buffer, 6);
}

uint8_t amsClearAux() {
  uint8_t buffer[6];
  return writeCMD(CLRAUX, buffer, 0);
}

uint8_t amsClearCells() {
  uint8_t buffer[6];
  return writeCMD(CLRCELL, buffer, 0);
}

uint8_t amsReadCellVoltages(Cell_Module* module) {
  uint8_t rxbuffer[CV_GROUP_A_SIZE];
  CHECK_RETURN(readCMD(RDCVA, rxbuffer, CV_GROUP_A_SIZE));
  module->cellVoltages[0] = mV_from_ADBMS6830(rxbuffer[0] | (rxbuffer[1] << 8));
  module->cellVoltages[1] = mV_from_ADBMS6830(rxbuffer[2] | (rxbuffer[3] << 8));
  module->cellVoltages[2] = mV_from_ADBMS6830(rxbuffer[4] | (rxbuffer[5] << 8));

  CHECK_RETURN(readCMD(RDCVB, rxbuffer, CV_GROUP_A_SIZE));
  module->cellVoltages[3] = mV_from_ADBMS6830(rxbuffer[0] | (rxbuffer[1] << 8));
  module->cellVoltages[4] = mV_from_ADBMS6830(rxbuffer[2] | (rxbuffer[3] << 8));
  module->cellVoltages[5] = mV_from_ADBMS6830(rxbuffer[4] | (rxbuffer[5] << 8));

  CHECK_RETURN(readCMD(RDCVC, rxbuffer, CV_GROUP_A_SIZE));
  module->cellVoltages[6] = mV_from_ADBMS6830(rxbuffer[0] | (rxbuffer[1] << 8));
  module->cellVoltages[7] = mV_from_ADBMS6830(rxbuffer[2] | (rxbuffer[3] << 8));
  module->cellVoltages[8] = mV_from_ADBMS6830(rxbuffer[4] | (rxbuffer[5] << 8));

  CHECK_RETURN(readCMD(RDCVD, rxbuffer, CV_GROUP_A_SIZE));
  module->cellVoltages[9]  = mV_from_ADBMS6830(rxbuffer[0] | (rxbuffer[1] << 8));
  module->cellVoltages[10] = mV_from_ADBMS6830(rxbuffer[2] | (rxbuffer[3] << 8));
  module->cellVoltages[11] = mV_from_ADBMS6830(rxbuffer[4] | (rxbuffer[5] << 8));

  CHECK_RETURN(readCMD(RDCVE, rxbuffer, CV_GROUP_A_SIZE));
  module->cellVoltages[12] = mV_from_ADBMS6830(rxbuffer[0] | (rxbuffer[1] << 8));
  module->cellVoltages[13] = mV_from_ADBMS6830(rxbuffer[2] | (rxbuffer[3] << 8));
  module->cellVoltages[14] = mV_from_ADBMS6830(rxbuffer[4] | (rxbuffer[5] << 8));

  CHECK_RETURN(readCMD(RDCVF, rxbuffer, CV_GROUP_A_SIZE));
  module->cellVoltages[15] = mV_from_ADBMS6830(rxbuffer[0] | (rxbuffer[1] << 8));

  return 0;
}
