/*
 * ADBMS_LL_Driver.c
 *
 *  Created on: 05.06.2022
 *      Author: max
 */

#include "ADBMS_LL_Driver.h"
#include "ADBMS_CMD_MAKROS.h"
#include "config_ADBMS6830.h"
#include "stm32h7xx_hal.h"
#include "swo_log.h"
#include <stdbool.h>
#include <stdint.h>
#include <strings.h>

#define INITIAL_COMMAND_PEC        0x0010
#define INITIAL_DATA_PEC           0x0010 
#define ADBMS_SPI_TIMEOUT          100 // Timeout in ms
#warning ask about the timeout value

SPI_HandleTypeDef* adbmsspi;
uint8_t command_queue[N_BMS][12] = {0};

uint8_t adbmsDriverInit(SPI_HandleTypeDef* hspi) {
  mcuAdbmsCSLow();
  HAL_Delay(1);
  mcuAdbmsCSHigh();
  adbmsspi = hspi;
  return 0;
}

static HAL_StatusTypeDef mcuSPITransmit(uint8_t* buffer, uint8_t buffersize) {
    HAL_StatusTypeDef status;
    status = HAL_SPI_Transmit(adbmsspi, buffer, buffersize, ADBMS_SPI_TIMEOUT);
    __HAL_SPI_CLEAR_OVRFLAG(adbmsspi);
    return status;
}

static HAL_StatusTypeDef mcuSPIReceive(uint8_t* buffer, uint8_t buffersize) {
    return HAL_SPI_Receive(adbmsspi, buffer, buffersize, ADBMS_SPI_TIMEOUT);
}

static HAL_StatusTypeDef mcuSPITransmitReceive(uint8_t* rxbuffer, uint8_t* txbuffer, uint8_t buffersize) {
    return HAL_SPI_TransmitReceive(adbmsspi, txbuffer, rxbuffer, buffersize, ADBMS_SPI_TIMEOUT);
}

//command PEC calculation
//CRC-15
//x^15 + x^14 + x^10 + x^8 + x^7 + x^4 + x^3 + 1

static uint16_t updateCommandPEC(uint16_t currentPEC, uint8_t din) {
  din = (din >> 7) & 0x01;
  uint8_t in0 = din ^ ((currentPEC >> 14) & 0x01);
  uint8_t in3 = in0 ^ ((currentPEC >> 2) & 0x01);
  uint8_t in4 = in0 ^ ((currentPEC >> 3) & 0x01);
  uint8_t in7 = in0 ^ ((currentPEC >> 6) & 0x01);
  uint8_t in8 = in0 ^ ((currentPEC >> 7) & 0x01);
  uint8_t in10 = in0 ^ ((currentPEC >> 9) & 0x01);
  uint8_t in14 = in0 ^ ((currentPEC >> 13) & 0x01);

  uint16_t newPEC = 0;

  newPEC |= in14 << 14;
  newPEC |= (currentPEC & (0x01 << 12)) << 1;
  newPEC |= (currentPEC & (0x01 << 11)) << 1;
  newPEC |= (currentPEC & (0x01 << 10)) << 1;
  newPEC |= in10 << 10;
  newPEC |= (currentPEC & (0x01 << 8)) << 1;
  newPEC |= in8 << 8;
  newPEC |= in7 << 7;
  newPEC |= (currentPEC & (0x01 << 5)) << 1;
  newPEC |= (currentPEC & (0x01 << 4)) << 1;
  newPEC |= in4 << 4;
  newPEC |= in3 << 3;
  newPEC |= (currentPEC & (0x01 << 1)) << 1;
  newPEC |= (currentPEC & (0x01)) << 1;
  newPEC |= in0;

  return newPEC;
}

static uint8_t calculateCommandPEC(uint8_t* data, uint8_t datalen) {
  uint16_t currentpec = INITIAL_COMMAND_PEC;
  if (datalen >= 3) {
    for (int i = 0; i < (datalen - 2); i++) {
      for (int n = 0; n < 8; n++) {
        uint8_t din = data[i] << (n);
        currentpec = updateCommandPEC(currentpec, din);
      }
    }

    data[datalen - 2] = (currentpec >> 7) & 0xFF;
    data[datalen - 1] = (currentpec << 1) & 0xFF;
    return 0;
  } else {
    return 1;
  }
}

static uint8_t checkCommandPEC(uint8_t* data, uint8_t datalen) {
  if (datalen <= 3) {
    return 255;
  }

  uint16_t currentpec = INITIAL_COMMAND_PEC;

  for (int i = 0; i < (datalen - 2); i++) {
    for (int n = 0; n < 8; n++) {
      uint8_t din = data[i] << (n);
      currentpec = updateCommandPEC(currentpec, din);
    }
  }

  uint8_t pechigh = (currentpec >> 7) & 0xFF;
  uint8_t peclow = (currentpec << 1) & 0xFF;

  if ((pechigh == data[datalen - 2]) && (peclow == data[datalen - 1])) {
    return 0;
  }

  return 1;
}

//data PEC calculation
//CRC-10
//x^10 + x^7 + x^3 + x^2 + x + 1

static uint16_t pec10_calc(bool rx_cmd, int len, uint8_t* data) {
  uint16_t remainder = 16; /* PEC_SEED;   0000010000 */
  uint16_t polynom = 0x8F; /* x10 + x7 + x3 + x2 + x + 1 <- the CRC15 polynomial
                              100 1000 1111   48F */

  /* Perform modulo-2 division, a byte at a time. */
  for (uint8_t pbyte = 0; pbyte < len; ++pbyte) {
    /* Bring the next byte into the remainder. */
    remainder ^= (uint16_t)(data[pbyte] << 2);
    /* Perform modulo-2 division, a bit at a time.*/
    for (uint8_t bit_ = 8; bit_ > 0; --bit_) {
      /* Try to divide the current data bit. */
      if ((remainder & 0x200) >
          0) // equivalent to remainder & 2^14 simply check for MSB
      {
        remainder = (uint16_t)((remainder << 1));
        remainder = (uint16_t)(remainder ^ polynom);
      } else {
        remainder = (uint16_t)(remainder << 1);
      }
    }
  }
  if (rx_cmd == true) {
    remainder ^= (uint16_t)((data[len] & 0xFC) << 2);
    /* Perform modulo-2 division, a bit at a time */
    for (uint8_t bit_ = 6; bit_ > 0; --bit_) {
      /* Try to divide the current data bit */
      if ((remainder & 0x200) >
          0) // equivalent to remainder & 2^14 simply check for MSB
      {
        remainder = (uint16_t)((remainder << 1));
        remainder = (uint16_t)(remainder ^ polynom);
      } else {
        remainder = (uint16_t)((remainder << 1));
      }
    }
  }
  return ((uint16_t)(remainder & 0x3FF));
}

typedef uint16_t crc;
static crc F_CRC_CalculaCheckSum(uint8_t const AF_Datos[], uint16_t VF_nBytes);

static uint8_t calculateDataPEC(uint8_t* data, uint8_t datalen) {

  if (datalen >= 3) {
    

    crc currentpec = pec10_calc(true, datalen - 2, data) & 0x3FF; // mask to 10 bits

    // memory layout is [[zeroes], PEC[9:8]], [PEC[7:0]]
    data[datalen - 2] = (currentpec >> 8) & 0xFF;
    data[datalen - 1] = currentpec & 0xFF;

    volatile uint8_t result = pec10_calc(true, datalen, data);

    return 0;
  } else {
    return 1;
  }
}

static uint8_t checkDataPEC(uint8_t* data, uint8_t len) {
  if (len <= 2) {
    return 255;
  }

  crc currentpec = F_CRC_CalculaCheckSum(data, len);

  return (currentpec == 0) ? 0 : 1;
}


static crc F_CRC_ObtenValorDeTabla(uint8_t VP_Pos_Tabla) {
  crc VP_CRCTableValue = 0;
  uint8_t VP_Pos_bit = 0;

  VP_CRCTableValue = ((crc)(VP_Pos_Tabla)) << (10 - 8);

  for (VP_Pos_bit = 0; VP_Pos_bit < 8; VP_Pos_bit++) {
    if (VP_CRCTableValue & (((crc)1) << (10 - 1))) {
      VP_CRCTableValue = (VP_CRCTableValue << 1) ^ 0x8F;
    } else {
      VP_CRCTableValue = (VP_CRCTableValue << 1);
    }
  }
  return ((VP_CRCTableValue));
}
static crc F_CRC_CalculaCheckSum(uint8_t const AF_Datos[], uint16_t VF_nBytes) {
  crc VP_CRCTableValue = 16;
  int16_t VP_bytes = 0;

  for (VP_bytes = 0; VP_bytes < VF_nBytes; VP_bytes++) {

    VP_CRCTableValue = (VP_CRCTableValue << 8) ^
                       F_CRC_ObtenValorDeTabla(
                           ((uint8_t)((VP_CRCTableValue >> (10 - 8)) & 0xFF)) ^
                           AF_Datos[VP_bytes]);
  }

  if ((8 * sizeof(crc)) > 10) {
    VP_CRCTableValue = VP_CRCTableValue & ((((crc)(1)) << 10) - 1);
  }

  return (VP_CRCTableValue ^ 0x0000);
}

static uint16_t updateDataPEC(uint16_t currentPEC, uint8_t din) {
  din = (din >> 7) & 0x01;
  uint8_t in0 = din ^ ((currentPEC >> 9) & 0x01);
  uint8_t in2 = in0 ^ ((currentPEC >> 1) & 0x01);
  uint8_t in3 = in0 ^ ((currentPEC >> 2) & 0x01);
  uint8_t in7 = in0 ^ ((currentPEC >> 6) & 0x01);

  uint16_t newPEC = 0;

  newPEC |= (currentPEC & (0x01 << 8)) << 1;
  newPEC |= (currentPEC & (0x01 << 7)) << 1;
  newPEC |= in7 << 7;
  newPEC |= (currentPEC & (0x01 << 5)) << 1;
  newPEC |= (currentPEC & (0x01 << 4)) << 1;
  newPEC |= in3 << 3;
  newPEC |= in2 << 2;
  newPEC |= (currentPEC & (0x01)) << 1;
  newPEC |= in0;

  return newPEC;
}
  
HAL_StatusTypeDef ___writeCMD(uint16_t command, uint8_t * args, size_t arglen) {
  HAL_StatusTypeDef ret = HAL_OK;
  if (arglen > 0) {
    args[0] = (command >> 8) & 0xFF;
    args[1] = (command) & 0xFF;

    if (DEBUG_CHANNEL_ENABLED(LOG_LEVEL_NOISY)) {
      debug_log(LOG_LEVEL_NOISY, "%lu W | %x %x ", HAL_GetTick(), args[0], args[1]);

      //print out data bytes
      for (size_t i = 0; i < N_BMS; i++) {
        debug_log_cont(LOG_LEVEL_NOISY, "%d: ", i);
        for (size_t j = 0; j < arglen; j++) {
          debug_log_cont(LOG_LEVEL_NOISY, "%x ", args[4 + (i * (arglen + 2)) + j]);
        }
      }

      debug_log_cont(LOG_LEVEL_NOISY, "\n");
    }

    calculateCommandPEC(args, 4);

    for (size_t i = 0; i < N_BMS; i++) {
      calculateDataPEC(&args[4 + (i * (arglen + 2))], arglen + 2); //DPEC is calculated over the data, not the command, and placed at the end of the data
    }
    
    mcuAdbmsCSLow();
    ret = mcuSPITransmit(args, CMD_BUFFER_SIZE(arglen));
    mcuAdbmsCSHigh();
  } else {
    args[0] = (command >> 8) & 0xFF;
    args[1] = (command) & 0xFF;
    calculateCommandPEC(args, 4);

    mcuAdbmsCSLow();
    ret = mcuSPITransmit(args, 4);
    mcuAdbmsCSHigh();
  }

  return ret;
}



HAL_StatusTypeDef readCMD(uint16_t command, uint8_t * buffer, size_t buflen) {
  buffer[0] = (command >> 8) & 0xFF;
  buffer[1] = (command)&0xFF;
  calculateCommandPEC(buffer, 4);

  mcuAdbmsCSLow();
  HAL_StatusTypeDef status = mcuSPITransmitReceive(buffer, buffer, buflen);
  mcuAdbmsCSHigh();

  if (status != HAL_OK) return status;

  //[[maybe_unused]] uint8_t commandCounter = buffer[sizeof(buffer) - 2] & 0xFC; //command counter is bits 7-2
                                                                                 //TODO: check command counter?

  if (DEBUG_CHANNEL_ENABLED(LOG_LEVEL_NOISY)) {
    debug_log(LOG_LEVEL_NOISY, "%lu R | %x %x ", HAL_GetTick(), command >> 8, command & 0xFF);

    //print out data bytes
    for (size_t i = 0; i < N_BMS; i++) {
      debug_log_cont(LOG_LEVEL_NOISY, "%d: ", i);
      for (size_t j = 0; j < buflen; j++) {
        debug_log_cont(LOG_LEVEL_NOISY, "%x ", buffer[4 + (i * (buflen + 2)) + j]);
      }
    }

    debug_log_cont(LOG_LEVEL_NOISY, "\n");
  }
                                                                                 
  return checkDataPEC(&buffer[4], buflen + 2);
}

//check poll command - no data PEC sent back
uint8_t pollCMD(uint16_t command) {
  uint8_t txbuffer[5] = {};
  uint8_t rxbuffer[5] = {};

  txbuffer[0] = (command >> 8) & 0xFF;
  txbuffer[1] = (command)&0xFF;
  calculateCommandPEC(txbuffer, 4);

  mcuAdbmsCSLow();
  uint8_t status = mcuSPITransmitReceive(rxbuffer, txbuffer, 5);
  mcuAdbmsCSHigh();

  if (status != 0) {
    return status;
  }

  return rxbuffer[4]; //last byte will be poll response
}

void mcuAdbmsCSLow() {
  //TODO: set right GPIO
  //HAL_GPIO_WritePin(CSB_GPIO_Port, CSB_Pin, GPIO_PIN_RESET);
}

void mcuAdbmsCSHigh() {
  //HAL_GPIO_WritePin(CSB_GPIO_Port, CSB_Pin, GPIO_PIN_SET);
}
